/*
 Illinois Open Source License
 
 University of Illinois/NCSA
 Open Source License
 
 Copyright © 2009,    University of Illinois.  All rights reserved.
 
 Developed by: 
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing Applications, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
 
 THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 */

#include "model.h"
#include "nallatech.h"
#include "network.h"

// an ifndef for the BenNUEY-PCI104-v4, as this is only in the latest dimesdl.h
#ifndef mbtTHEBENNUEYPCI104V4     /* Check this is defined */
    #define mbtTHEBENNUEYPCI104V4 0x0000000C
#endif
#ifndef mbtH100PCIX
    #define mbtH100PCIX 0x00000011
#endif

/**********************************************/
/* Converts from a slot number to a JTAG      */
/* module number                              */
/**********************************************/
int SlotToModule(DIME_HANDLE hCard, int Slot)
{
    int slotsused = DIME_CardStatus(hCard,dinfSLOTSUSED);
    int nummodules = DIME_CardStatus(hCard,dinfNUMBERMODULES);
    int numslots = DIME_CardStatus(hCard,dinfNUMBERSLOTS);
    int cardtype = DIME_CardStatus(hCard,dinfMOTHERBOARDTYPE);
    int retval=-1;
    int i;

    printf("slotsused : %d, nummodules %d, numslots: %d\n",slotsused, nummodules, numslots);

    // adjustment for BenNUEY-PCIX or BenNUEY-PCI104-v4
    if((cardtype==mbtTHEBENNUEYPCIX) || (cardtype==mbtTHEBENNUEYPCI104V4)) {
        numslots = numslots-1;
        slotsused = slotsused-16;

        // Check if MicroDime slot is used & if so reduce nummodules by 1
        // MicroDime has the highest module number
        if (slotsused & (1<<3))
          nummodules--;
    }

 
    if (Slot==-1)
      retval = nummodules-1;
    else
    {
      if (slotsused & (1<<Slot)) {
        retval = 0;
        for (i=numslots-1;i>Slot;i--) {
          if (slotsused & (1<<i))
            retval++;
        }
      }
    }

    if (retval==-1) {
      if (Slot>=numslots)
        printf("Error : Slot %d does not exist on the card\n",Slot);
      else
        printf("Error : Slot %d is not populated\n",Slot);
    }
    printf("Slot %d: Module %d\n", Slot, retval);
    return(retval);
}

/**********************************************/
/* Takes the DIMEtalk Capability return and   */
/* renders the information to a readable form */
/**********************************************/
void DisplayCapability(int node, unsigned int capability[2])
{
  char typenames[6][20] = {"RAM","Memory Map","VHDL Node","Source FIFO","Sink FIFO",
                        "Processor"};
  char procname[4][20] = {"picoblaze","microblaze","PowerPC","x86 PC"};
  int type = capability[0]&0xff;
  int nodeid = (capability[0]&0xff00)>>8;
  int subtype = (capability[0]&0xff0000)>>16;
  int version = (capability[0]&0xff000000)>>24;
  int rcapable = (capability[1]&0x1);
  int wcapable = (capability[1]&0x2)>>1;
  int acapable = (capability[1]&0x4)>>2;
  int mdoorbell = (capability[1]&0x8)>>3;
  int wdoorbell = ((capability[1]&0x30)>>4)+mdoorbell;
  int rdoorbell = ((capability[1]&0xc0)>>6)+mdoorbell;
  int status = (capability[1]&0xff000000)>>24;
  char subtypedesc[30]="\0";

  switch (type) {
    case 0 : sprintf(subtypedesc,"RAM size = %d unsigned ints",1<<(subtype+9));break;
    case 3 :
    case 4 : sprintf(subtypedesc,"FIFO size = %d unsigned ints",(1<<(subtype+9))-1);break;
    case 5 : sprintf(subtypedesc,"%s",procname[subtype]);break;
  }

  printf("\n");
  printf("  Node Address    : %d\n",node);
  printf("    Status        : 0x%x\n",status);
  printf("    Node Ident    : %d\n",nodeid);
  printf("    Node Version  : %d\n",version);
  printf("    Type          : %s (%d)\n",typenames[type],type);
  printf("    Subtype       : %s\n",subtypedesc);
  printf("    Capabilities\n");
  printf("      Readable    : %s\n",(rcapable?"Yes":"No"));
  printf("      Writeable   : %s\n",(wcapable?"Yes":"No"));
  printf("      Atomic Ops  : %s\n",(acapable?"Yes":"No"));
  printf("    Doorbells\n");
  printf("      # Write     : %d\n",wdoorbell);
  printf("      # Read      : %d\n",rdoorbell);
}

/*********************************************/
/* Simple Test Code.                         */
/* Assumes FIFOs and Memory Map nodes have   */
/* Loopbacks attached                        */
/*********************************************/
int Node_Test(DIMETALK_HANDLE hTalk, int nodenumber, unsigned int Capability[2])
{
  int size=0;
  //unsigned int *outdata=NULL; //Changed from DWORD
  DWORD *outdata=NULL; //Changed from DWORD
  //unsigned int *indata=NULL; //Changed from DWORD
  DWORD *indata=NULL; //Changed from DWORD
  int i;
  int type = Capability[0]&0xff;
  int subtype = (Capability[0]&0xff0000)>>16;

  switch (type) {
    case 0 : size = 1<<(subtype+9);break;
  }

  if (!size) {
    printf("Node of type %d is currently not tested\n",type);
    return(0);
  }

  outdata = calloc(size,4);
  if (!outdata)
  {
    printf("outdata: Memory allocation failed.\n");
    return(1);
  }

  indata =  calloc(size,4);
  if (!indata)
  {
    printf("indata: Memory allocation failed.\n");
    free(outdata);
    return(1);
  }

  printf("Node test. Node = %d. Size = %d\n",nodenumber,size);

  for (i=0;i<size;i++) {
    outdata[i] = i | (i<<16);
  }
  if (DIMETalk_Write(hTalk,outdata,size,0,nodenumber,1000))
  {
      printf("Error during write. Node %d. Type %d\n",nodenumber,type);
      free(indata);
      free(outdata);
      return(1);
  }
  if (DIMETalk_Read(hTalk,indata,size,0,nodenumber,1000))
  {
      printf("Error during read. Node %d. Type %d\n",nodenumber,type);
      free(indata);
      free(outdata);
      return(1);
  }
  for (i=0;i<size;i++) {
    if (outdata[i] != indata[i]) {
      printf("Error during comparison. Word %d. Sent %li. Got %li. Node %d. Type %d\n",
             i,outdata[i],indata[i],nodenumber,type);
      free(indata);
      free(outdata);
      return(1);
    }
  }
  free(indata);
  free(outdata);
  printf("Test complete...\n\n\n");
  return(0);
}

/*********************************************/
/* Main Code                                 */
/*********************************************/
struct accelerator* nallatech_init()
{
    struct accelerator* myacc = NULL;

    LOCATE_HANDLE hLocate = NULL;
    unsigned int /*ErrorNum,*/ NumOfCards,LoopCntr,ConfigResult,CardType;
    DWORD ErrorNum;
    char ErrorString[1000];
    DIME_HANDLE hCard;
    DIMETALK_HANDLE hTalk;
    unsigned int Capability[2];
    double ActualFreq;
    int Module, Count;
    SWMBInfo *pBoardInfo;

    // The iftype variable below specifies the hardware interface, substitute with dlUSB, dlVME etc as required.
    unsigned int iftype = dlPCI;

    //printf("DIMETalk API Example Code...\n\n");

// *****DIMEtalk: User Code Start
// *****DIMEtalk: User Code End

    pBoardInfo = (SWMBInfo *)DIME_SystemStatusPtr(0,dinfDIME_SWMBTS);

    // Call the function to locate all Nallatech cards on the PCI interface.
    if( (hLocate = DIME_LocateCard(iftype,mbtALL,NULL,dldrDEFAULT,dlDEFAULT)) == NULL)
    {// Error hLocate NULL
            // Print the error then terminate the program
            DIME_GetError(NULL,&ErrorNum,ErrorString);
            printf("Error Number %li\n", ErrorNum);
            printf("%s\n",ErrorString);
            //printf ("\nPress return to terminate the application.\n");
            //getchar();
            //exit(1);
            return NULL;
    }

    // Determine how many Nallatech cards have been found.
    NumOfCards = DIME_LocateStatus(hLocate,0,dlNUMCARDS);
    printf("%d Nallatech card(s) found.\n", NumOfCards);

    // Get the details for each card detected.
    for (LoopCntr=1; LoopCntr<=NumOfCards; LoopCntr++)
    {
            unsigned int type = DIME_LocateStatus(hLocate,LoopCntr,dlMBTYPE);
            unsigned int typecount, found;
            printf("Details of card number %d, of %d:\n",LoopCntr,NumOfCards);
            typecount=found=0;
            while (typecount<pBoardInfo->NumTypes && !found) {
              if (pBoardInfo->pCardInfo[typecount].MotherBoardType==type) {
                found=1;
                printf("\tCard Description : %s\n",pBoardInfo->pCardInfo[typecount].MotherBoardDesc);
              }
              typecount++;
            }
            if (!found) {
                printf("\tThis is an unrecognised card. Driver %s\n",(char*)DIME_LocateStatusPtr(hLocate,LoopCntr,dlDESCRIPTION));
            }
    }

    // Open up the first card found. To open the nth card found simple change the second argument to n.
    hCard = DIME_OpenCard(hLocate,1,dccOPEN_DEFAULT); //opens up card 1 with default flags
    if (hCard == NULL) //check to see if the open worked.
    {
            printf("Card Number One failed to open.\n");
            DIME_CloseLocate(hLocate);
            //printf ("\nPress return to terminate the application.\n");
            //getchar();
            return NULL;
    }

    // Determine what type of card type it is.
    CardType = DIME_CardStatus(hCard,dinfMOTHERBOARDTYPE);

    // If it's a BenNUEY-PCIX or a BenNUEY-PCI104-v4 we need to set the oscilators to programmable
    if((CardType==mbtTHEBENNUEYPCIX) || (CardType==mbtTHEBENNUEYPCI104V4)){
      printf("Setting the clocks to the programmable oscillators\n");
      DIME_OscillatorControl(hCard,1,oscOSCSELECT,oscPROGOSC);
      DIME_OscillatorControl(hCard,2,oscOSCSELECT,oscPROGOSC);
      DIME_OscillatorControl(hCard,3,oscOSCSELECT,oscPROGOSC);
    }

    // Set up the oscilators frequencies
    #ifdef CLKA_FREQUENCY
    DIME_SetOscillatorFrequency(hCard,1,CLKA_FREQUENCY,&ActualFreq);
    #else
    DIME_SetOscillatorFrequency(hCard,1,100.0,&ActualFreq);
    #endif
    printf("Clock A set to %f\n",ActualFreq);
    #ifdef CLKB_FREQUENCY
    DIME_SetOscillatorFrequency(hCard,2,CLKB_FREQUENCY,&ActualFreq);
    #else
    DIME_SetOscillatorFrequency(hCard,2,40.0,&ActualFreq);
    #endif
    printf("Clock B set to %f\n",ActualFreq);
    #ifdef CLKC_FREQUENCY
    DIME_SetOscillatorFrequency(hCard,3,CLKC_FREQUENCY,&ActualFreq);
    #else
    DIME_SetOscillatorFrequency(hCard,3,100.0,&ActualFreq);
	#endif
    printf("Clock C set to %f\n",ActualFreq);

    printf("Starting reset and configuration sequence...\n");

    // Check to see if the onboard reset is supported on the particular card.
    if (CardType >= 0x10)
    {
        if (DIME_CardResetStatus(hCard,drONBOARDFPGA,drONBOARDFPGA)==drCONTROLABLE)
          DIME_CardResetControl(hCard,drONBOARDFPGA, drENABLE,0);
    }
    else if (CardType != 0xB)   // check it's not a BenONE-104
        DIME_CardResetControl(hCard,drONBOARDFPGA, drENABLE,0);


    // enable the ststem reset
    DIME_CardResetControl(hCard,drSYSTEM, drENABLE,0);

    // disable the system reset
    DIME_CardResetControl(hCard,drSYSTEM, drDISABLE,0);
    //Check to see if the onboard fpga reset is supported on the particular card.
    if (CardType >= 0x10)
    {
        if (DIME_CardResetStatus(hCard,drONBOARDFPGA,drONBOARDFPGA)==drCONTROLABLE)
          DIME_CardResetControl(hCard,drONBOARDFPGA, drDISABLE,0);
    }
    else if (CardType != 0xB)   // check it's not a BenONE-104
          DIME_CardResetControl(hCard,drONBOARDFPGA, drDISABLE,0);


    // toggle the interface reset
    DIME_CardResetControl(hCard,drINTERFACE, drTOGGLE,0);


   //Check to see if the onboard fpga reset is supported on the particular card.
    if (CardType >= 0x10)
    {
       if (DIME_CardResetStatus(hCard,drONBOARDFPGA,drONBOARDFPGA)==drCONTROLABLE)
          DIME_CardResetControl(hCard,drONBOARDFPGA, drENABLE,0);
    }
    else if (CardType != 0xB)   // check it's not a BenONE-104
         DIME_CardResetControl(hCard,drONBOARDFPGA, drENABLE,0);

    // enable the ststem reset
    DIME_CardResetControl(hCard,drSYSTEM, drENABLE,0);

    // Load the devices in the network
    printf("Loading network design files...\n");
    if (CardType != mbtH100PCIX)
    {
        for (Count=0;Count<NUMDEVICES;Count++)
        {
          unsigned int Module = SlotToModule(hCard,deviceinfo[Count][0]);
          printf("%s...",devicefiles[Count]);
          ConfigResult = DIME_ConfigDevice(hCard,devicefiles[Count],Module,deviceinfo[Count][1],0,0);
          printf("done, with configresult = %x\n", ConfigResult);
        }
    }
    else
    {
          printf("%s...",devicefiles[0]);
          ConfigResult = DIME_ConfigDevice(hCard,devicefiles[0],(DIME_CardStatus(hCard,dinfNUMBERMODULES)-1),0,0,0);
          printf("done, with configresult = %x\n", ConfigResult);
    }


    //Disable the resets
   //Check to see if the onboard fpga reset is supported on the particular card.
    if (CardType >= 0x10)
    {
       if (DIME_CardResetStatus(hCard,drONBOARDFPGA,drONBOARDFPGA)==drCONTROLABLE)
          DIME_CardResetControl(hCard,drONBOARDFPGA, drDISABLE,0);
    }
    else if (CardType != 0xB)   // check it's not a BenONE-104
          DIME_CardResetControl(hCard,drONBOARDFPGA, drDISABLE,0);

    DIME_CardResetControl(hCard,drSYSTEM, drDISABLE,0);
    Sleep(50);
    //Enable the resets
   //Check to see if the onboard fpga reset is supported on the particular card.
    if (CardType >= 0x10)
    {
       if (DIME_CardResetStatus(hCard,drONBOARDFPGA,drONBOARDFPGA)==drCONTROLABLE)
          DIME_CardResetControl(hCard,drONBOARDFPGA, drENABLE,0);
    }
    else if (CardType != 0xB)   // check it's not a BenONE-104
          DIME_CardResetControl(hCard,drONBOARDFPGA, drENABLE,0);


    DIME_CardResetControl(hCard,drSYSTEM, drENABLE,0);
    Sleep(50);
    //Disable the resets
   //Check to see if the onboard fpga reset is supported on the particular card.
    if (CardType >= 0x10)
    {
       if (DIME_CardResetStatus(hCard,drONBOARDFPGA,drONBOARDFPGA)==drCONTROLABLE)
          DIME_CardResetControl(hCard,drONBOARDFPGA, drDISABLE,0);
    }
    else if (CardType != 0xB)   // check it's not a BenONE-104
          DIME_CardResetControl(hCard,drONBOARDFPGA, drDISABLE,0);


    DIME_CardResetControl(hCard,drSYSTEM, drDISABLE,0);
    DIME_CardResetControl(hCard,drINTERFACE, drTOGGLE,0);
    Sleep(50);

    printf("Finished reset and configuration sequence...\n");

    // Open DIMEtalk API. Replace 0 with 1 to open using interrupts for DMAs
    //printf("About to call DIMETalk_Open\n");
    hTalk = DIMETalk_Open(hCard,0);

/*
    for (Count=0;Count<NUMNODES;Count++)
    {
      printf("For Node %d\n", nodeaddr[Count]);
      DIMETalk_Capability(hTalk,Capability,nodeaddr[Count],1000);
      printf("Got capability back\n");
      printf("Returned: 0x%lx  0x%lx\n",Capability[0],Capability[1]);
      DisplayCapability(nodeaddr[Count],Capability);
      // Test nodes. This assumes that FIFO and memory map nodes have loopbacks connected.
      Node_Test(hTalk,nodeaddr[Count],Capability);
    }
*/
    myacc = (struct accelerator *)malloc(sizeof(struct accelerator));
    myacc->hTalk = hTalk;
    myacc->hLocate = hLocate;
    myacc->hCard = hCard;

    return myacc;
}


void nallatech_close(struct accelerator *myacc)
{
    DIMETalk_Close(myacc->hTalk);  // Closes DIMEtalk API
    DIME_CloseCard(myacc->hCard); // Closes down the card.

    // Finally the last thing that should be done is to close down the locate.
    DIME_CloseLocate(myacc->hLocate);

    //printf ("\nPress return to terminate the application.\n");
    //getchar();

    free(myacc);

    return ;
}

// *****DIMEtalk: User Code Start

void doComputeFPGA(struct accelerator *myacc,
                  struct cartesian *data1, int n1, 
                  struct cartesian *data2, int n2,
                  int doSelf, 
                  long long **data_bins, int nbins, double *binb, int njk)
{
    //struct accelerator *myacc = nallatech_init();

    int n1n2, i, j, timeout = 1000;

    double *sram0 = (double *)malloc((n1 + n2) * sizeof(double));
    double *sram1 = (double *)malloc((n1 + n2) * sizeof(double));
    double *sram2 = (double *)malloc((n1 + n2) * sizeof(double));
    double *sram3 = (double *)malloc((n1+njk*(nbins + 2)) * sizeof(double));

    if (doSelf)
    {
        n2 = n1;
        data2 = data1;
    }

    for (i = 0; i < n1; i++)
    {
       sram0[i] = data1[i].x;
       sram1[i] = data1[i].y;
       sram2[i] = data1[i].z;
       sram3[i] = data1[i].jk;
    }
    n1n2 = n1;

    if (!doSelf)
    {
        for (i = 0; i < n2; i++)
        {
           sram0[i+n1] = data2[i].x;
           sram1[i+n1] = data2[i].y;
           sram2[i+n1] = data2[i].z;
        }
        n1n2 += n2;
    }

    for (i = 0; i < nbins+1; i++)
    {
        sram3[n1+i] = binb[i];
    }

    // send scalar values
    int data32[5];
    data32[0] = n1;
    data32[1] = n2;
    data32[2] = nbins;
    data32[3] = njk;
    data32[4] = doSelf;  

    if (DIMETalk_Write(myacc->hTalk, (DWORD*)data32, 5, 0+1, memory_map_0, timeout))
    {
        printf("Error during DIMETalk_Write.\n");
        exit(1);
    }

    // send arrays
    printf("  Sending %li bytes of data to SRAMs...", ((nbins+1)+(n1n2)*3+n1)*sizeof(double));

    if (DIMETalk_Write(myacc->hTalk, (DWORD*)sram0, 2*(n1n2), 0, ddr2sram_if_h100_0, timeout))
    {
        printf("Error during DIMETalk_Write.\n");
        exit(1);
    }
    if (DIMETalk_Write(myacc->hTalk, (DWORD*)sram1, 2*(n1n2), 0, ddr2sram_if_h100_1, timeout))
    {
        printf("Error during DIMETalk_Write.\n");
        exit(1);
    }
    if (DIMETalk_Write(myacc->hTalk, (DWORD*)sram2, 2*(n1n2), 0, ddr2sram_if_h100_2, timeout))
    {
        printf("Error during DIMETalk_Write.\n");
        exit(1);
    }

    if (DIMETalk_Write(myacc->hTalk, (DWORD*)sram3, 2*(n1+nbins+1), 0, ddr2sram_if_h100_3, timeout))
    {
        printf("Error during DIMETalk_Write.\n");
        exit(1);
    }

    printf(" done!\n");

    // run it

    printf("  Waiting for FPGA to return...\n");

    DWORD Control = 1;
    DIMETalk_Read(myacc->hTalk, &Control, 1, 0, memory_map_0, timeout);
    if (Control&0x1) Control = 0; else Control = 1;
    DIMETalk_Write(myacc->hTalk, &Control, 1, 0, memory_map_0, timeout);

    while (!(Control&0x2))
      DIMETalk_Read(myacc->hTalk, &Control, 1, 0, memory_map_0, timeout);

    printf("  Receiving %li bytes of data... ", njk*(nbins+2)*sizeof(double));

    // copy results from the card
    if (DIMETalk_Read(myacc->hTalk, (DWORD*)sram3, njk*2*(nbins+2), 0, ddr2sram_if_h100_3, timeout))
    { 
        printf("Error during DIMETalk_Read.\n");
        exit(1);
    }

    printf("done!\n");

    for (j = 0; j < njk; j++)
    for (i = 0; i < nbins+2; i++)
    {
        //data_bins[0][i] += (int)((doSelf) ? ((i==0) ? 0 : sram3[i]/2) : sram3[i]);
        data_bins[j][i] += (doSelf && i == 0) ? sram3[(nbins+2)*j+i] - n1/2 : sram3[(nbins+2)*j+i];
        //printf("%i: out: %.12f\n", i, sram3[i]);
    }

    free(sram0);
    free(sram1);
    free(sram2);
    free(sram3);

    //nallatech_close(myacc);

    return;
}

// *****DIMEtalk: User Code End



